"""Command line interface for all modules"""
import argparse
import os
import sys

from forester import folder_contributions, tree_info


def show_version():
    print()


def main(*raw_args):
    # Main parser
    parser = argparse.ArgumentParser(description="Swiss-army knife for directory trees")
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Suppress unnecessary output, such as progress reports",
    )
    parser.add_argument(
        "--version", help="Show version information",
    )
    subparsers = parser.add_subparsers()

    # Parser for subcommand "info"
    info_parser = subparsers.add_parser(
        "info", help="Informative directory tree overview"
    )
    info_parser.add_argument(
        "path", default=os.getcwd(), nargs="?", help="Directory to analyze"
    )
    info_parser.set_defaults(func=tree_info.print_tree_info)

    # Parser for subcommand "contribs"
    contribs_parser = subparsers.add_parser(
        "contribs",
        help="List the contributions of subfolders to the total disk space",
    )
    contribs_parser.add_argument(
        "path", default=os.getcwd(), nargs="?", help="Directory to analyze"
    )
    contribs_parser.set_defaults(func=folder_contributions.print_folder_contributions)


    # Common path argument
    args = parser.parse_args(raw_args)

    if args.version:
        show_version()
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_usage()


def call_main():
    main(*sys.argv[1:])


if __name__ == "__main__":
    call_main()



#!/usr/bin/python
"""List the total disk space and disk space contributions of all subfolders in the given path.

Both figures only differ if there are inodes which are hardlinks from multiple of these folders. Then the contributions
show how much disk space would be gained if one of the folders would be deleted. Inodes with hardlinks from outside
would not be deleted in this case.
"""
import os
import stat
import sys

from forester.util import format_number


def folder_contributions(folder_path, verbose):
    """Get the total disk space and disk space contributions of all subfolders in the given path.

    Returns:
        Tuple of dictionaries: (totals, contributions)
        Both are in the form {folder_name: size}
    """
    folder_path = str(folder_path)
    inodes_read = set()
    inode_sizes = {}

    def getsize(folder_path, recurse=True):
        assert os.path.isdir(folder_path)

        if verbose and len(inode_sizes) % 100 == 0:
            sys.stdout.write(f"\r... scanning ({len(inode_sizes)} folders) ...")
            sys.stdout.flush()

        # Collect data about the folder itself
        st = os.lstat(folder_path)
        uniqueinode = "%s%s" % (st[stat.ST_INO], st[stat.ST_DEV])
        size = st[stat.ST_SIZE]
        inode_sizes[uniqueinode] = size
        inodes_read.add(uniqueinode)
        inodes = {uniqueinode}

        # Collect data about the children
        for dir_entry in os.scandir(folder_path):
            if dir_entry.is_dir(follow_symlinks=False):
                if recurse:
                    s, i = getsize(dir_entry.path)
                    size += s
                    inodes.update(i)
            else:
                uniqueinode = "%s%s" % (
                    dir_entry.inode(),
                    dir_entry.stat(follow_symlinks=False)[stat.ST_DEV],
                )
                inodes.add(uniqueinode)
                if uniqueinode in inodes_read:
                    size += inode_sizes[uniqueinode]
                else:
                    size += dir_entry.stat(follow_symlinks=False)[stat.ST_SIZE]
                    inode_sizes[uniqueinode] = dir_entry.stat(follow_symlinks=False)[
                        stat.ST_SIZE
                    ]
                    inodes_read.add(uniqueinode)

        return size, inodes

    def calc_contribution(inodes, inode_sets):
        exclusive_inodes = inodes
        for s in inode_sets:
            exclusive_inodes = exclusive_inodes.difference(s)

        return sum([inode_sizes[i] for i in exclusive_inodes])

    folder_names = [
        f for f in os.listdir(folder_path) if os.path.isdir(folder_path + "/" + f) and f != '.'
    ]
    inode_sets = {}
    totals = {}
    for f in folder_names:
        s, i = getsize(folder_path + "/" + f)
        inode_sets[f] = i
        totals[f] = s

    if verbose:
        sys.stdout.write(f"\rScanned {len(inodes_read)} inodes                      \n")
        sys.stdout.flush()

    contributions = {}
    for f in folder_names:
        contributions[f] = calc_contribution(
            inode_sets[f], [inode_sets[other] for other in folder_names if other != f]
        )

    s, i = getsize(folder_path, recurse=False)
    totals["."] = sum(inode_sizes.values())
    contributions["."] = calc_contribution(i, inode_sets)

    return totals, contributions


def output(name, total, contrib, colsize):
    print(
        f"{name.ljust(colsize)}{format_number(total).rjust(colsize)}{format_number(contrib).rjust(colsize)}"
    )


def print_folder_contributions(args):
    colsize = 30
    sort_order = "contribs"

    p = os.path.realpath(args.path)

    totals, contributions = folder_contributions(p, verbose=(not args.quiet))

    output("Folder", "Total size (B)", "Size of unique inodes (B)", colsize)
    print("-" * (3 * colsize))

    if sort_order == "name":
        folders = sorted(totals.keys(), reverse=True)
    elif sort_order == "totals":
        folders = (k for k, v in sorted(totals.items(), key=lambda item: item[1]))
    elif sort_order == "contribs":
        folders = (
            k for k, v in sorted(contributions.items(), key=lambda item: item[1])
        )

    for f in folders:
        if f != ".":
            output(f, totals[f], contributions[f], colsize)

    print("=" * (3 * colsize))
    output("Total", totals["."], "-", colsize)



from datetime import datetime, timezone


def format_number(number) -> str:
    """Format a number with thousand separators and return a string"""
    try:
        return "{:,}".format(number)
    except ValueError:
        return str(number)


def format_timestamp(timestamp) -> str:
    """Format a unix timestamp in ISO date format"""
    try:
        return datetime.fromtimestamp(timestamp, tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
    except TypeError:
        return str(timestamp)


def cut_to_length(text: str, length: int):
    if len(text) > length:
        return f"{text[:length-3]}..."
    return text



#!/usr/bin/python
"""List basic information for all subfolders in the given path.

TODO: Add usage info
"""
from collections import namedtuple
import os
from os.path import realpath, relpath
import stat
import sys
from typing import Dict

from forester.util import cut_to_length, format_number, format_timestamp


TreeInfo = namedtuple("TreeInfo", ["file_count", "folder_count", "last_mtime"])


def tree_info(folder_path, verbose=False) -> Dict[str, TreeInfo]:
    """Get the total disk space and disk space contributions of all subfolders in the given path.

    Returns:
        Dictionary: (totals, contributions)
        Both are in the form {folder_name: size}
    """
    folder_path = str(folder_path)
    collected_info = {}

    def collect_folder_info(path: str) -> TreeInfo:
        assert os.path.isdir(path)

        # Display progress info
        if verbose and len(collected_info) % 100 == 0:
            sys.stdout.write(f"\r... scanning ({len(collected_info)} folders) ...")
            sys.stdout.flush()

        # Collect data about the folder itself
        st = os.lstat(path)
        max_m_time = st[stat.ST_MTIME]
        total_n_files = 0
        total_n_folders = 1

        # Collect data about the children
        for dir_entry in os.scandir(path):
            if dir_entry.is_dir(follow_symlinks=False):
                n_files, n_folders, m_time = collect_folder_info(dir_entry.path)
                total_n_files += n_files
                total_n_folders += n_folders
                max_m_time = max(max_m_time, m_time)
            else:
                max_m_time = max(
                    max_m_time, dir_entry.stat(follow_symlinks=False)[stat.ST_MTIME]
                )
                total_n_files += 1

        collected_info[path] = TreeInfo(
            total_n_files, total_n_folders, max_m_time
        )

        return collected_info[path]

    collect_folder_info(folder_path)
    if verbose:
        sys.stdout.write(f"\rScanned {len(collected_info)} folders           ")
        print("\n")
    return collected_info


def _output(path, *info_fields, column_widths):
    line = f"{cut_to_length(path, column_widths[0]-1).ljust(column_widths[0]-1)}"
    line += ''.join(
        f" {cut_to_length(f, column_widths[i+1]-1).rjust(column_widths[i+1]-1)}"
        for i, f in enumerate(info_fields)
    )
    print(line)


def print_tree_info(args):
    column_widths = (50, 15, 15, 20)

    path = realpath(args.path)

    collected_info: Dict[str, TreeInfo] = tree_info(path, verbose=(not args.quiet))

    _output(
        "Folder",
        "# Folders",
        "# Files",
        "m_time",
        column_widths=column_widths
    )
    print("-" * sum(column_widths))

    folders = sorted(filter(lambda k: realpath(k) != path, collected_info.keys()), reverse=True)
    folders.append(path)

    for f in folders:
        if f == path:
            print("-" * sum(column_widths))
        _output(
            relpath(f, path),
            format_number(collected_info[f].folder_count),
            format_number(collected_info[f].file_count),
            format_timestamp(collected_info[f].last_mtime),
            column_widths=column_widths
        )
